#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from typing import Optional, List
import argparse
import os
import sys
import tempfile
import shutil
import base64
from enum import Enum
import json

from jinja2 import Template

from .libs import Config, get_config, AppConfig, HostConfig

class ConfigDeployType(Enum):
    COPY        = "copy"
    CONVERT     = "convert"
    TEMPLATE    = "template"

def init_hosts(base_dir: str, config: Config, host_names: List[str]) -> None:
    host_dict = {}
    for host_name in host_names:
        host = config.get_host(host_name)
        if host is None:
            print(f"Host {host_name} does not exist.")
            sys.exit(1)
        else:
            host_dict[host_name] = host

    for host_name in host_names:
        init_host(base_dir, config, host_dict[host_name])


def init_host(base_dir: str, config: Config, host: HostConfig) -> None:
    """ Initialize a host for mordor

    :param base_dir:
    :param config:
    :param host:
    :return: Nothing
    """

    print(f"Initialize mordor for host {host.name} ...")
    for dir_name in [
        host.env_home,
        host.path("bin"),
        host.path("apps"),
        host.path("venvs"),
        host.path("pids"),
        host.path("logs"),
        host.path("configs"),
        host.path("data"),
        host.path("temp"),
    ]:
        host.execute("mkdir", "-p", dir_name)

    cmds = [
        "install_packages.sh",
        "run_dispatcher.sh",
        "init_host.sh",
        "host_tools.txt",
    ]
    for cmd in cmds:
        host.upload(
            os.path.join(base_dir, "bin", cmd),
            host.path("bin", cmd)
        )
        if cmd.endswith(".sh"):
            host.execute("chmod", "+x", host.path("bin", cmd))

    host.execute(host.path("bin", "init_host.sh"), host.env_home)
    print(f"Done!")


def stage_app(
    config: Config,
    app_name: str,
    update_venv: bool,
    config_only: bool,
    stage: str = '',
    host_names: Optional[List[str]] = None
) -> None:
    """ Stage an application on the fleet for a stage

    :param base_dir: name of the directory contains mordor.py
    :param config: overall config
    :param app_name: application name
    :param update_venv: create virtual environment or not
    :param config_only: update config only or not
    :param stage: application stage, e.g., "beta", "prod", etc.
    :param host_names: if specified, we only stage to this list of hosts, otherwise, we stage to all hosts for the stage
    :return: Nothing
    """
    app = config.get_app(app_name, stage)
    if app is None:
        print(f"Application {app_name} with stage {stage} does not exist.")
        sys.exit(1)
    if not app.use_python3:
        print(f"Application {app_name} with stage {stage} does not support python3.")
        sys.exit(1)

    # archive the entire app and send it to host
    # tar -czf /tmp/a.tar.gz -C $PWD *
    if not config_only:
        archive_filename = app.create_archive()
    else:
        archive_filename = None

    if host_names is not None:
        deploy_to = host_names
    else:
        deploy_to = app.deploy_to

    # do a check first
    for host_name in deploy_to:
        host = config.get_host(host_name)
        if host is None:
            print(f"Host {host_name} does not exist.")
            sys.exit(1)

    for host_name in deploy_to:
        host = config.get_host(host_name)
        stage_app_on_host(config, app, host, archive_filename, update_venv, config_only, stage=stage)


def stage_app_on_host(
    config: Config,
    app: AppConfig,
    host: HostConfig,
    archive_filename: str,
    update_venv: bool,
    config_only: bool,
    stage: str = ''
) -> None:
    """ Stage an application on target host

    :param config: overall config
    :param app: application config
    :param host: host config
    :param archive_filename: the application archive filename
    :param update_venv: create virtual environment or not
    :param config_only: update config only or not
    :param stage: application stage, e.g., "beta", "prod", etc.
    :return: Nothing
    """
    print(f"Stage application {app.name} for stage {app.stage} on host {host.name}.")
    # allow user to have different configs for different stages
    config_base_dir = os.path.join(config.config_dir, "configs", app.name)
    app_config_dirs = []

    def add_dir_to_config_dirs(dir_name: str) -> None:
        if os.path.isdir(dir_name) and dir_name not in app_config_dirs:
            app_config_dirs.append(dir_name)

    add_dir_to_config_dirs(os.path.join(config_base_dir, stage, host.name))
    add_dir_to_config_dirs(os.path.join(config_base_dir, stage))
    add_dir_to_config_dirs(config_base_dir)

    def find_config_filename(name: str) -> str:
        for prefix in app_config_dirs:
            full_name = os.path.join(prefix, name)
            if os.path.isfile(full_name):
                return full_name
        print(f"Config file {name} does not exist.")
        sys.exit(1)

    # creating local staging area for upload file via scp
    temp_dir = tempfile.mkdtemp()
    local_stage_dir = os.path.join(temp_dir, app.name)
    os.makedirs(local_stage_dir)
    if not config_only:
        shutil.copyfile(archive_filename, os.path.join(local_stage_dir, app.archive_filename))
    # generate metadata.json
    metadata_filename = os.path.join(local_stage_dir, "_deployment.json")
    with open(metadata_filename, "wt") as f:
        json.dump(
            {
                "host": host.to_json(),
                "app": app.to_json(),
            },
            f,
            indent=4
        )
    for (filename, deploy_type) in app.config.items():
        if deploy_type == ConfigDeployType.COPY.value:
            shutil.copyfile(
                find_config_filename(filename),
                os.path.join(local_stage_dir, filename)
            )
            continue
        if deploy_type == ConfigDeployType.CONVERT.value:
            with open(find_config_filename(filename), "r") as f:
                content = f.read()
            config_dir = os.path.join(host.env_home, "configs", app.name)
            content = content.format(
                config_dir=config_dir,
                env_home=host.env_home,
                app_name=app.name
            )
            with open(os.path.join(local_stage_dir, filename), "w") as sf:
                sf.write(content)
            continue
        if deploy_type == ConfigDeployType.TEMPLATE.value:
            with open(find_config_filename(filename), "rt") as f:
                template = Template(f.read())
                context = {
                    "host_name": host.name,
                    "config_dir": os.path.join(host.env_home, "configs", app.name),
                    "log_dir": os.path.join(host.env_home, "logs", app.name),
                    "data_dir": os.path.join(host.env_home, "data", app.name),
                    "pid_dir": os.path.join(host.env_home, "pids", app.name),
                    "env_home": host.env_home,
                    "app_name": app.name
                }
            with open(os.path.join(local_stage_dir, filename), "wt") as sf:
                sf.write(template.render(context))
            continue

    if not config_only:
        print("    Upload application and configuration ... ", end="", flush=True)
    else:
        print("    Upload configuration ... ", end="", flush=True)
    host.upload_batch(local_stage_dir, host.path("temp"))
    print("Done!")

    # copy app archive to remote host
    lines = []
    if not config_only:
        # create directories
        lines.extend([
            f"mkdir -p {host.path('apps', app.name)}",
            f"mkdir -p {host.path('apps', app.name, app.manifest.version)}",
            f"rm -rf {host.path('apps', app.name, app.manifest.version, '*')}",
            f"mkdir -p {host.path('logs', app.name)}",
            f"mkdir -p {host.path('configs', app.name)}",
            f"mkdir -p {host.path('data', app.name)}",
            f"mkdir -p {host.path('pids', app.name)}",
            # remove and re-create the sym link point to the current version of the app
            f"rm -f {host.path('apps', app.name, 'current')}",
            f"ln -s {host.path('apps', app.name, app.manifest.version)} {host.path('apps', app.name, 'current')}",
            # extract app archive
            f"tar -xzf {host.path('temp', app.name, app.archive_filename)} -C {host.path('apps', app.name, app.manifest.version)}",
        ])
    # move config file from temp dir
    for filename in list(app.config.keys()) + ['_deployment.json']:
        lines.append(f"mv {host.path('temp', app.name, filename)} {host.path('configs', app.name, filename)}")
    if not config_only:
        # finally, let's remove the archive_filename and staging area in temp
        lines.append(f"rm {host.path('temp', app.name, app.archive_filename)}")
    lines.append(f"rmdir {host.path('temp', app.name)}")

    if not config_only:
        # recreate venv since dependencies may have changed
        if update_venv:
            lines.append(f"rm -rf {host.path('venvs', app.venv_name)}")
            if app.use_python3:
                lines.append(f"{host.python3} -m venv {host.path('venvs', app.venv_name)}")
            else:
                lines.append(f"{host.virtualenv} {host.path('venvs', app.venv_name)}")
            lines.append(
                f"{host.path('bin', 'install_packages.sh')} {host.env_home} {app.name} {app.manifest.version} {app.requirements}")
            lines.append(f"rm -f {host.path('venvs', app.name)}")
            # create a symlink
            lines.append(f"ln -s {host.path('venvs', app.venv_name)} {host.path('venvs', app.name)}")
            print("    Update application, configuration and virtual environment ... ", end="", flush=True)
        else:
            print("    Update application and configuration ... ", end="", flush=True)
    else:
        print("    Update configuration ... ", end="", flush=True)

    host.execute_batch(lines)
    print("Done!")

    if app.manifest.on_stage is not None:
        run_app_on_host(app, host, app.manifest.on_stage, prefix="    ")
    print("Done!")
    print("")


def run_app(
    config: Config,
    app_name: str,
    stage:str = '',
    host_names: Optional[List[str]] = None,
    cmd:str = ""
) -> None:
    """Run an application on the fleet for a stage

    :param config: overall config
    :param app_name: application name
    :param stage: application stage, e.g., "beta", "prod", etc.
    :param host_names: if specified, we only stage to this list of hosts, otherwise, we stage to all hosts for the stage
    :param cmd: the command to run
    :return: Nothing
    """
    app = config.get_app(app_name, stage)
    if app is None:
        print(f"Application {app_name} with stage {stage} does not exist.")
        sys.exit(1)
    if not app.use_python3:
        print(f"Application {app_name} with stage {stage} does not support python3.")
        sys.exit(1)

    if host_names is not None:
        run_on = [host_names]
    else:
        run_on = app.deploy_to

    # do a check first
    for host_name in run_on:
        host = config.get_host(host_name)
        if host is None:
            print(f"Host {host_name} does not exist.")
            sys.exit(1)

    for host_name in run_on:
        host = config.get_host(host_name)
        run_app_on_host(app, host, cmd)


def run_app_on_host(
    app: AppConfig,
    host: HostConfig,
    cmd: str,
    prefix:str = ""
):
    """ Run an application on target host

    :param app: application config
    :param host: host config
    :param cmd: command to run
    :return: Nothing
    """
    print(f"{prefix}Running application {app.name} on host {host.name}, cmd: \"{cmd}\".")

    cmd_to_send = base64.b64encode(cmd.encode('utf-8')).decode('utf-8')

    host.execute(
        host.path("bin", "run_dispatcher.sh"),
        host.env_home,
        app.name,
        cmd_to_send
    )

def main() -> None:
    parser = argparse.ArgumentParser(
        description='Mordor deployment tool for python.'
    )
    parser.add_argument(
        "action", type=str, help="Specify action",
        choices=['init-host', 'stage', 'run'],
        nargs=1
    )
    parser.add_argument(
        "-o", "--host-names", type=str, default=None, required=False, nargs='+',
        help="destination hosts"
    )
    parser.add_argument(
        "-p", "--app-name", type=str, required=False, help="application name"
    )
    parser.add_argument(
        "-s", "--stage", type=str, required=False, help="stage"
    )
    parser.add_argument(
        "-cmd", "--cmd", type=str, required=False, help="command to run",
        default=""
    )
    parser.add_argument(
        "--update-venv",
        action="store_true",
        help="Specify if you want to update virtualenv",
    )
    parser.add_argument(
        "--config-only",
        action="store_true",
        help="Specify if you want to stage configuration only",
    )
    parser.add_argument(
        "-c", "--config-dir", type=str, required=False, help="Configuration directory",
        default=os.environ.get("MORDOR_CONFIG_DIR", os.path.expanduser("~/.mordor"))
    )
    args = parser.parse_args()

    action = args.action[0]

    # validate options
    if args.update_venv and args.config_only:
        print("You cannot specify both --update-venv and --config-only")
        sys.exit(1)

    base_dir = os.path.abspath(os.path.dirname(__file__))

    # get the config file
    config_dir = os.path.expanduser(args.config_dir)
    for filename in [
        os.path.join(args.config_dir, 'config.yaml'),
        os.path.join(args.config_dir, 'config.json'),
        None
    ]:
        if filename is not None and os.path.isfile(filename):
            break
    if filename is None:
        print(f"Missing config file in directory {args.config_dir}")
        sys.exit(1)

    config = Config(get_config(filename), config_dir)

    if action == "init-host":
        if not args.host_names:
            print("--host-names must be specified.")
            sys.exit(1)
        init_hosts(base_dir, config, args.host_names)
        return

    if action == "stage":
        if not args.app_name:
            print("--app-name must be specified.")
            sys.exit(1)
        stage_app(
            config, args.app_name, args.update_venv, args.config_only,
            stage=args.stage,
            host_names=args.host_names
        )
        return

    if action == "run":
        run_app(
            config, args.app_name,
            stage = args.stage,
            host_names = args.host_names,
            cmd = args.cmd
        )
        return

if __name__ == '__main__':
    main()



import os
import json
import yaml


def get_json(path: str):
    with open(os.path.expanduser(path), "r") as f:
        return json.load(f)


def get_yaml(path: str):
    with open(os.path.expanduser(path), "r") as f:
        return yaml.safe_load(f)


def get_config(path: str):
    if path.endswith(".json"):
        return get_json(path)
    if path.endswith(".yaml"):
        return get_yaml(path)
    assert False, "Impossible"



from typing import Dict, List
import subprocess
import os
import glob
import tempfile
import platform

from .app_manifest import AppManifest
from .tools import get_config


class AppConfig:
    """Represent configuration for a given deployment
    """

    deployment_name: str  # the name of the deployment
    app_config: dict

    def __init__(self, deployment_name: str, app_config: dict):
        self.app_config = app_config
        self.deployment_name = deployment_name

        self.manifest = None
        for manifest_filename in [
            self.path("manifest.yaml"),
            self.path("manifest.json"),
        ]:
            if os.path.isfile(manifest_filename):
                self.manifest = AppManifest(get_config(manifest_filename))
                break
        if self.manifest is None:
            raise Exception("Missing manifest file")

    @property
    def name(self) -> str:
        return self.app_config.get("name", self.deployment_name)

    @property
    def stage(self) -> str:
        return self.app_config.get("stage", "")

    @property
    def home_dir(self) -> str:
        return self.app_config["home_dir"]

    @property
    def requirements(self) -> str:
        return self.app_config.get('requirements', 'requirements.txt')

    @property
    def use_python3(self) -> bool:
        return self.app_config.get("use_python3", True)

    @property
    def config(self) -> Dict[str, str]:
        # config files need to copied over
        return self.app_config.get("config", {})

    @property
    def deploy_to(self) -> List[str]:
        return self.app_config["deploy_to"]

    def path(self, *args) -> str:
        return os.path.join(self.home_dir, *args)

    @property
    def archive_filename(self) -> str:
        return "{}-{}.tar.gz".format(self.name, self.manifest.version)

    @property
    def venv_name(self) -> str:
        return "{}-{}".format(self.name, self.manifest.version)

    def create_archive(self) -> str:
        temp_dir = tempfile.mkdtemp()
        exclude_opts = []
        for exclude_dir in self.manifest.exclude_dirs:
            exclude_opts.extend(["--exclude", exclude_dir])

        # macos might generate ._files, use --disable-copyfile to disable it
        args = ['tar', '--no-xattrs']
        if platform.system() == "Darwin":
            args.append("--disable-copyfile")
        args += exclude_opts + \
            [
                '-czf',
                os.path.join(temp_dir, self.archive_filename),
                "-C",
                self.home_dir
            ]
        files_to_add = glob.glob("{}/*".format(self.home_dir))
        files_to_add = [item[len(self.home_dir) + 1:] for item in files_to_add]
        args.extend(files_to_add)
        subprocess.check_call(args)
        return os.path.join(temp_dir, self.archive_filename)

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "stage": self.stage,
            "home_dir": self.home_dir,
            "deploy_to": self.deploy_to,
            "manifest": self.manifest.to_json()
        }



import subprocess
import os
from typing import Optional, List, Tuple
import tempfile


class HostConfig:
    """Represent configuration for a given host
    """

    name: str  # the unique name of the host we address this host, not necessary be the DNS name
    host_config: dict  # the config dictionary for this host

    def __init__(self, host_name: str, host_config: dict):
        self.host_config = host_config
        self.name = host_name

    @property
    def env_home(self) -> str:
        return self.host_config["env_home"]

    @property
    def ssh_host(self) -> str:
        return self.host_config.get("ssh_host", self.name)

    @property
    def python3(self) -> Optional[str]:
        return self.host_config.get("python3", "/usr/bin/python3")

    def path(self, *args) -> str:
        return os.path.join(self.env_home, *args)

    def execute_batch(self, lines: List[str]) -> None:
        """ Execute a bash batch file on this host

        :param lines: lines for the batch script
        :return: Nothing
        """
        with tempfile.NamedTemporaryFile(delete=True) as f:
            for line in lines:
                f.write(f"{line}\n".encode("utf-8"))
            f.write(f"exit\n".encode("utf-8"))
            f.seek(0)

            new_args = ["ssh", "-q", self.ssh_host]
            subprocess.check_call(new_args, stdin=f)

    def execute(self, *args) -> None:
        """ Execute a one-line command on this host

        :param args: args for this script
        :return: Nothing
        """
        new_args = ["ssh", "-q", "-t", self.ssh_host]
        new_args.extend(args)
        subprocess.check_call(new_args)

    def execute2(self, *args) -> Tuple[str, int]:
        """ Execute a script on this host, capture the output.

        :param args: args for this script
        :return: tuple, script stdout as string, and script exit code as integer
        """
        new_args = ["ssh", self.ssh_host]
        new_args.extend(args)
        p = subprocess.Popen(new_args, stdout=subprocess.PIPE)
        _ = p.wait()
        output, err = p.communicate()
        return output, err

    def upload_batch(self, local_path: str, remote_path: str) -> None:
        """ Upload an entire directory to the host

        :param local_path: the path to the directory which need upload
        :param remote_path: the destination path
        :return: Nothing
        """
        scp_options = os.environ.get('SCP_OPTIONS','').strip()
        options = scp_options.split(" ") if scp_options else []
        new_args = ["scp"] + options + ["-r", "-q",
            local_path,
            "{}:{}".format(self.ssh_host, remote_path)
        ]
        subprocess.check_call(new_args)

    def upload(self, local_path: str, remote_path: str) -> None:
        """ Upload a file to the host

        :param local_path: the path to the file which need upload
        :param remote_path: the destination path
        :return: Nothing
        """
        scp_options = os.environ.get('SCP_OPTIONS','').strip()
        options = scp_options.split(" ") if scp_options else []
        new_args = ["scp"] + options + ["-q",
            local_path,
            "{}:{}".format(self.ssh_host, remote_path)
        ]
        subprocess.check_call(new_args)

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "env_home": self.env_home,
            "ssh_host": self.ssh_host,
        }



import os
import json
import yaml
import glob
import shutil
import importlib
from copy import deepcopy
from typing import Optional
from jinja2 import Template
from .libs.tools import get_config
from .libs.app_manifest import AppManifest

class AppEnv:
    def __init__(self, app_name: str, version:Optional[str] = None, env_home:Optional[str] = None):
        self.env_home = os.environ["ENV_HOME"] if env_home is None else env_home
        self.app_name = app_name
        self._manifest = None

        if version is None:
            app_base_dir = os.path.join(self.env_home, "apps", self.app_name, "current")
            manifest_filenames = [
                os.path.join(app_base_dir, "manifest.yaml"),
                os.path.join(app_base_dir, "manifest.json"),
            ]
            app_manifest = None
            for manifest_filename in manifest_filenames:
                if os.path.isfile(manifest_filename):
                    app_manifest = AppManifest(get_config(manifest_filename))
                    break
            if app_manifest is None:
                raise Exception("Missing manifest")
            self.version = app_manifest.version
        else:
            self.version = version

    @property
    def app_dir(self) -> str:
        return os.path.join(self.env_home, "apps", self.app_name, self.version)

    @property
    def log_dir(self) -> str:
        return os.path.join(self.env_home, "logs", self.app_name)

    @property
    def data_dir(self) -> str:
        return os.path.join(self.env_home, "data", self.app_name)

    @property
    def config_dir(self) -> str:
        return os.path.join(self.env_home, "configs", self.app_name)

    @property
    def pid_dir(self) -> str:
        return os.path.join(self.env_home, "pids", self.app_name)

    @property
    def venv_dir(self) -> str:
        return os.path.join(self.env_home, "venvs", f"{self.app_name}-{self.version}")

    def get_json_config(self, filename: str):
        full_path = os.path.join(
            self.env_home, "configs", self.app_name, filename
        )
        with open(full_path, "r") as f:
            return json.load(f)

    def get_yaml_config(self, filename: str):
        full_path = os.path.join(
            self.env_home, "configs", self.app_name, filename
        )
        with open(full_path, "r") as f:
            return yaml.safe_load(f)

    def get_config(self, filename: str):
        if filename.endswith(".json"):
            return self.get_json_config(filename)
        if filename.endswith(".yaml"):
            return self.get_yaml_config(filename)
        assert False, "Only json or yaml are supported"

def deltree(p:str) -> None:
    """ Delete everything inside a directory

    :param p: directory location
    :return: Nothing
    """
    files = []
    files.extend(glob.glob(os.path.join(p, ".*"), recursive=False))
    files.extend(glob.glob(os.path.join(p, "*"), recursive=False))
    for file in files:
        if os.path.isfile(file):
            os.remove(file)
        elif os.path.isdir(file):
            shutil.rmtree(file)


def prepare_for_docker(app_name: str, context: dict = {}) -> None:
    app_env = AppEnv(app_name)
    _context = app_env.get_config("_deployment.json")
    _context.update(context)
    process_templates(app_name, "docker", _context)
    app_docker_dir = os.path.join(app_env.data_dir, "docker", "app")
    if os.path.isdir(app_docker_dir):
        shutil.rmtree(app_docker_dir)
    def to_ignore(dir_name, contents):
        if dir_name == app_env.app_dir:
            return ['docker']
        else:
            return []
    shutil.copytree(
        os.path.join(app_env.app_dir),
        app_docker_dir,
        symlinks=True,
        ignore=to_ignore
    )


def process_templates(app_name: str, template_dir: str, context: dict) -> None:
    """ Process template directory, result stored in data dir

    :param app_name: application name
    :param template_dir: the relative path name of the template directory, relative to the app home
    :param context: the context we need to bind to
    :return: Nothing
    """

    app_env = AppEnv(app_name)
    src_dir = os.path.join(app_env.app_dir)
    dst_dir = os.path.join(app_env.data_dir)

    dst_template_dir = os.path.join(dst_dir, template_dir)
    os.makedirs(dst_template_dir, exist_ok=True)
    deltree(dst_template_dir)

    process_templates_dir(src_dir, dst_dir, template_dir, context)

def process_templates_dir(src_base_dir:str, dst_base_dir:str, template_dir: str, context: dict):
    src_dir = os.path.join(src_base_dir, template_dir)
    dst_dir = os.path.join(dst_base_dir, template_dir)

    if os.path.isfile(os.path.join(src_dir, "context.py")):
        module = importlib.import_module(
            ".".join(list(os.path.split(template_dir))[1:] + ['context'])
        )
        get_context = getattr(module, 'get_context')
        local_context = get_context(deepcopy(context))
    else:
        local_context = context

    for file in os.listdir(src_dir):
        if file in ("context.py", "__pycache__"):
            continue
        src_filename = os.path.join(src_dir, file)

        if os.path.isfile(src_filename):
            if src_filename.endswith(".template"):
                with open(src_filename, "rt") as rf:
                    template = Template(rf.read())
                    dst_filename = os.path.join(dst_dir, file[:-9]) # remove .template
                    with open(dst_filename, "wt") as wf:
                        wf.write(template.render(local_context))
            else:
                dst_filename = os.path.join(dst_dir, file)
                shutil.copyfile(src_filename, dst_filename)
        elif os.path.isdir(src_filename):
            sub_template_dir = os.path.join(template_dir, file)
            os.makedirs(os.path.join(dst_base_dir, sub_template_dir), exist_ok=False)
            process_templates_dir(src_base_dir, dst_base_dir, sub_template_dir, local_context)
